knockoutDebugCallback([
    'src/namespace.js',
    'src/google-closure-compiler-utils.js',
    'src/version.js',
    'src/options.js',
    'src/utils.js',
    'src/utils.domData.js',
    'src/utils.domNodeDisposal.js',
    'src/utils.domManipulation.js',
    'src/memoization.js',
    'src/tasks.js',
    'src/subscribables/extenders.js',
    'src/subscribables/subscribable.js',
    'src/subscribables/dependencyDetection.js',
    'src/subscribables/observable.js',
    'src/subscribables/observableArray.js',
    'src/subscribables/observableArray.changeTracking.js',
    'src/subscribables/dependentObservable.js',
    'src/subscribables/mappingHelpers.js',
    'src/subscribables/observableUtils.js',
    'src/binding/selectExtensions.js',
    'src/binding/expressionRewriting.js',
    'src/virtualElements.js',
    'src/binding/bindingProvider.js',
    'src/binding/bindingAttributeSyntax.js',
    'src/components/loaderRegistry.js',
    'src/components/defaultLoader.js',
    'src/components/customElements.js',
    'src/components/componentBinding.js',
    'src/binding/defaultBindings/attr.js',
    'src/binding/defaultBindings/checked.js',
    'src/binding/defaultBindings/css.js',
    'src/binding/defaultBindings/enableDisable.js',
    'src/binding/defaultBindings/event.js',
    'src/binding/defaultBindings/foreach.js',
    'src/binding/defaultBindings/hasfocus.js',
    'src/binding/defaultBindings/html.js',
    'src/binding/defaultBindings/ifIfnotWith.js',
    'src/binding/defaultBindings/let.js',
    'src/binding/defaultBindings/options.js',
    'src/binding/defaultBindings/selectedOptions.js',
    'src/binding/defaultBindings/style.js',
    'src/binding/defaultBindings/submit.js',
    'src/binding/defaultBindings/text.js',
    'src/binding/defaultBindings/textInput.js',
    'src/binding/defaultBindings/uniqueName.js',
    'src/binding/defaultBindings/using.js',
    'src/binding/defaultBindings/value.js',
    'src/binding/defaultBindings/visibleHidden.js',
    // click depends on event - The order matters for specs, which includes each file individually
    'src/binding/defaultBindings/click.js',
    'src/templating/templateEngine.js',
    'src/templating/templateRewriting.js',
    'src/templating/templateSources.js',
    'src/templating/templating.js',
    'src/binding/editDetection/compareArrays.js',
    'src/binding/editDetection/arrayToDomNodeChildren.js',
    'src/templating/native/nativeTemplateEngine.js',
    'src/templating/jquery.tmpl/jqueryTmplTemplateEngine.js'
]);
